
#include "general.h"
#include "allocate.h"

#include "function.h"
#include "function.icc"

#include "GridMap.h"
//#include "GridMap.icc"

#include "P_Motion.h"

#include <string>
using namespace std;

#ifdef WIN32
#pragma warning (disable : 4786 )
#endif

//The angles from center for each of the sensors.
static float SensorAngle[8] = {50.0, 31.0, 10.0, -10.0, -31.0, -50.0, -163.0, 163.0}; 
//The different rotations of the 8 sensors, I got these from the AmigoBot spec.
static float SensorRot[8] = {90.0, 44.0, 12.0, -12.0, -44.0, -90.0, -144.0, 144.0 };
//The distance from center (in mm) for each of the sensors
static float    SensorDist[8]= {120.0, 150.0, 160.0, 160.0, 150.0, 120.0, 160.0, 160.0};

static float util_GetDeviation( float left, float center, float right );

///////////
/////////////////////////////////////////////////////////////////
////////////////////////////////////////////////////////////////////////////
//
// Grid maps.
//
////////////////////////////////////////////////////////////////////////////
////////////////////////////////////////////////////////////////////////////


//----------------------------------------------------------------------
// Initialize
//----------------------------------------------------------------------

//----------------------------------------------------------------------
// 
//----------------------------------------------------------------------
GridMap::GridMap( string mapFile ) :
  _initialized( false),
  _distance_table_initialized(false),
  _sizeX( 0),
  _sizeY( 0),
  _prob( NULL),
  _resolution( DEFAULT_RESOLUTION),
  _offsetX( 0.0),
  _offsetY( 0.0),
  _maxRealX( 0.0),
  _maxRealY( 0.0),
  _unknownProb( DEFAULT_UNKNOWN_PROB),
  _fileName(mapFile),
  _distanceResolution(DEFAULT_DISTANCE_RES_LOOKUP)
{

  if ( ! mapFile.empty())
    read( mapFile);
  _onlineLookup = precomputedLookupTables();
}

//----------------------------------------------------------------------
// 
//----------------------------------------------------------------------
GridMap::GridMap( string mapFile, string distFile ) :
  _initialized( false),
  _distance_table_initialized(false),
  _sizeX( 0),
  _sizeY( 0),
  _prob( NULL),
  _resolution( DEFAULT_RESOLUTION),
  _offsetX( 0.0),
  _offsetY( 0.0),
  _maxRealX( 0.0),
  _maxRealY( 0.0),
  _unknownProb( DEFAULT_UNKNOWN_PROB),
  _fileName(mapFile),
  _distanceResolution(DEFAULT_DISTANCE_RES_LOOKUP)
{

  if ( ! mapFile.empty())
    read( mapFile);
  if ( ! distFile.empty())
    readExpectedDistanceTable(distFile);

  _onlineLookup = precomputedLookupTables();
}

GridMap::~GridMap()
{

}

//----------------------------------------------------------------------
// Reads the map from a file.
//----------------------------------------------------------------------
bool
GridMap::read( string& mapFile)
{
  register int x,y;
  float temp;
  
  char line[255];
  FILE *fp;

  //cerr << "# Read grid map from file " << mapFile << endl;
  _fileName = mapFile;
  
  // Free previous map if necesary.
  free2D( _prob, _sizeX);

  if ((fp = fopen(mapFile.c_str(),"rt")) == NULL) {
    fprintf(stderr,"ERROR: Could not open file %s\n", mapFile.c_str());
    exit(0);
  }
  
  _resolution = DEFAULT_RESOLUTION;
  
  /* Seek header for map data in rhino-Type file,
     analyze robot specifications */
  while ((fgets(line,255,fp) != NULL)
         && (strncmp("global_map[0]", line , 13) != 0)){
    
    if (strncmp(line,"robot_specifications->resolution",32) == 0) {
      if (sscanf(&line[32],"%f",&(_resolution)) != 0 ) {
	if ( _resolution <= 0.0) {
	  cerr << "Wrong resolution: %f.\n", _resolution;
	  exit(0);
	}
	else
	  fprintf( stderr,"# Map resolution: %f cm\n",_resolution);
	
      }
    }
    if (strncmp(line,"robot_specifications->autoshifted_x",35) == 0) {
      if (sscanf(&line[35],"%g",&(_offsetX)) != 0 ) {
	if ( _offsetX > 0.0)
	  fprintf( stderr, "# Map offsetX: %g cm\n", _offsetX);
      }
    }
    if (strncmp(line,"robot_specifications->autoshifted_y",35) == 0) {
      if (sscanf(&line[35],"%g",&(_offsetY)) != 0 ) {
	if ( _offsetY > 0.0)
	  fprintf( stderr, "# Map offsetY: %g cm\n", _offsetY);
      }
    }
  }
  if (sscanf (line,"global_map[0]: %d %d",&_sizeY,&_sizeX)
      == EOF) {
    fprintf(stderr,"ERROR: corrupted file %s\n",mapFile.c_str());
    fclose(fp);
    return(0);
  }
  
  
  fprintf( stderr,"# Map size: %d %d\n",_sizeX,_sizeY);
  
  _unknownProb = DEFAULT_UNKNOWN_PROB;
  _offsetX = _offsetY = 0;

  allocate2D( & _prob, _sizeX, _sizeY);
  
  if (_prob == NULL){
    fprintf(stderr,
              "ERROR: Not enough memory for loading map %s\n",mapFile.c_str());
    fclose(fp);
    return(0);
  }

  for (x=0;x < _sizeX; x++)
    for (y=0;y< _sizeY; y++) {
      fscanf(fp,"%e",&temp);
      if (temp < 0.0 || temp > 1.0)
	_prob[x][y] = _unknownProb;
      else 
	_prob[x][y] = 1-temp;	   
    }

  // Now find individual unexplored pixels.
  for (x=1; x < _sizeX - 1; x++)
    for (y=1; y < _sizeY - 1; y++) {
      if (_prob[x][y] == _unknownProb) {
	int cnt = 0;
	float sum = 0.0;
	if (_prob[x][y-1] != _unknownProb) {
	  cnt++;
	  sum += _prob[x][y-1];
	}
	if (_prob[x][y+1] != _unknownProb) {
	  cnt++;
	  sum += _prob[x][y+1];
	}
	if (_prob[x+1][y] != _unknownProb) {
	  cnt++;
	  sum += _prob[x+1][y];
	}
	if (_prob[x-1][y] != _unknownProb) {
	  cnt++;
	  sum += _prob[x-1][y];
	}

	if ( cnt > 2) {
	  //fprintf( stderr, "# Found isolated cell at %d %d.\n", x, y);
	  _prob[x][y] = sum / cnt;
	}
      }
    }
      
  fclose(fp);

  _maxRealX = _offsetX + _sizeX * _resolution;
  _maxRealY = _offsetY + _sizeY * _resolution;
  
  _initialized = true;

  double displayMaxX = _sizeX * MILLIMETERS_PER_CENTIMETER;
  double displayMaxY = _sizeY * MILLIMETERS_PER_CENTIMETER;

  return(1);
}

double GridMap::mapToDisplay( double mapVal )
{
  double ret = (((mapVal * _resolution) / CENTIMETERS_PER_PIXEL ) * MILLIMETERS_PER_CENTIMETER);
  return ret;
}

double GridMap::displayToMap( double dispVal )
{
  double ret = ((dispVal * CENTIMETERS_PER_PIXEL) / (_resolution * MILLIMETERS_PER_CENTIMETER) );
  return ret;
}

void GridMap::draw(SfWin *w )
{
  //now cycle through the _prob array and display the different probabilities as
  //rectangles

  //NOTICE!  Because of the limited capabilities of Saphira, the rectangles will be
  //drawn only if the probability for the given grid cell exceeds P_OBSTACLE_THRESHOLD

  for( int xval= 0; xval< (int)_sizeX; xval++)
  {
    for( int yval = 0; yval < (int)_sizeY; yval++)
    {
       double val = _prob[xval][yval];
       if( _prob[xval][yval] > OCCUPIED_THRESHOLD || _prob[xval][yval] == -1 )
       {  
          w->Rectangle(mapToDisplay(xval), mapToDisplay(yval), mapToDisplay(xval+1), mapToDisplay(yval+1));
       }
    }
  }
  
  //draw the beams from the current robot to the expected endpoints
  double botX = SfROBOT->getX();
  double botY = SfROBOT->getY();
  double botThDeg = SfROBOT->getTh();
  double botO = deg2Rad( botThDeg );

  State start( botX, botY, botO );
  PolarMotion p;
  State sensorPoint;
  State endPoint;

  for( int i=0; i<8; ++i )
  {
    // now draw a line for each of the sensors...
    p.set(SensorDist[i], deg2Rad(SensorAngle[i]),deg2Rad(SensorRot[i]));
    sensorPoint = p.endPoint(start);
    double sensorX = sensorPoint.value(X_DIM);
    double sensorY = sensorPoint.value(Y_DIM);
    double sensorRot = sensorPoint.value(ROT_DIM);
    int sensorDeg = rad2Deg(sensorRot); //debug friendly
    double distance = expectedDistance( botX, botY, botThDeg, i).first;

    //is this really working?
    p.set( distance, deg2Rad(0.0), deg2Rad(0.0));
    endPoint = p.endPoint( sensorPoint );
    double endX = endPoint.value(X_DIM);
    double endY = endPoint.value(Y_DIM);

    /*
    //do it yourself approach...
    double sinth = sin( sensorRot );
    double costh = cos( sensorRot );

    double yoffset = sinth * distance;
    double xoffset = costh * distance;

    double endX = sensorX + xoffset;
    double endY = sensorY + yoffset;
    */

    w->Vector( sensorX, sensorY, endX, endY);
  }
  
}

expectedDistanceRet
GridMap::expectedDistance( int botXmm, int botYmm, float botTh, int sensorNo )
{
  bool good_bad(true);

  float xval, yval, orientation;

  //first, find the position of the sensor
  PolarMotion p( SensorDist[sensorNo], deg2Rad(SensorAngle[sensorNo]), deg2Rad(SensorRot[sensorNo]) );
  State start( botXmm, botYmm, deg2Rad(botTh) );
  State endPoint = p.endPoint( start );
  float sensorXmm = endPoint.value(X_DIM);
  float sensorYmm = endPoint.value(Y_DIM);
  float sensorTh = endPoint.value(ROT_DIM);
  
  //now convert to cm, doing this painstakingly so I can debug it sanely
  float sensorXcm = (sensorXmm / MILLIMETERS_PER_CENTIMETER );
  float sensorYcm = (sensorYmm / MILLIMETERS_PER_CENTIMETER );

  //finally, get the grid position
  indexGridPosition indexPosition = indexPos( sensorXcm, sensorYcm );

  //these are the values we'll do the lookup on
  xval = indexPosition.x;
  yval = indexPosition.y;
  orientation = sensorTh;

    
  if (!inMap( xval, yval))
       return make_pair(0.0,false);

  int index;


  if( !_distance_table_initialized)
  {
    //now get the expectedDistanceIndex, multiply by the distanceResolution to get an answer in cm.
    index = expectedDistanceIndex( xval, yval, orientation, _onlineLookup );

    float orientation_rt_deg = rad2Deg(orientation)+(CONE_SIZE_DEGREES/2);
    float orientation_lft_deg = rad2Deg(orientation)-(CONE_SIZE_DEGREES/2);

    int index_rt = expectedDistanceIndex( xval, yval, deg2Rad(orientation_rt_deg), _onlineLookup );
    int index_lft = expectedDistanceIndex( xval, yval, deg2Rad(orientation_lft_deg), _onlineLookup );

    float variance = util_GetDeviation( index_lft, index, index_rt);

    if( variance >= DEVIATION_THRESHOLD_CONE )
            good_bad = false;

    index *= _distanceResolution;
  }
  else
  {
    int lookup_angle = (int)rad2Deg(orientation);
    int xidx = (int)xval;
    int yidx = (int)yval;

    index = _expectedDistTable[xidx][yidx][lookup_angle];
    if( index > 127 )
    {
      good_bad = false;
      index-=128;
    }
    index *= _distanceResolution;
  }

  //return an answer in mm, since that's what Amigobots care about.
  return make_pair( index * MILLIMETERS_PER_CENTIMETER, good_bad );
}

//----------------------------------------------------------------------
// Precomputes the lookup tables for fast retrieval of the expected distances.
//----------------------------------------------------------------------
distanceLookup
GridMap::precomputedLookupTables( float distanceResolution, int maxDistanceIndex )
{  
  distanceLookup lookupTable;

  lookupTable.distanceResolution = distanceResolution;
  _distanceResolution = distanceResolution;
  lookupTable.maxDistanceIndex   = maxDistanceIndex;
  lookupTable.angleToIndex       = ANGLE_TO_INDEX;
  lookupTable.numberOfAngles     = NUMBER_OF_OFFSET_ANGLES;

  //fprintf( stderr, "# Precompute lookup table: %f %d ... ", distanceResolution, maxDistanceIndex);
  
  float angleStep = DEG_360 / NUMBER_OF_OFFSET_ANGLES;
  int maxNumberOfDifferentSteps = (int) (2 * maxDistanceIndex * distanceResolution / _resolution + 1);
  
  // lookupTable.xOffset, lookupTable.yOffset :
  // Offset in the real map used to get the i-th index along the z-th beam.
  allocate2D( & lookupTable.xOffset, NUMBER_OF_OFFSET_ANGLES,
	      maxNumberOfDifferentSteps);
  
  allocate2D( & lookupTable.yOffset, NUMBER_OF_OFFSET_ANGLES,
	      maxNumberOfDifferentSteps);
  
  // lookupTable.numberOfOffsets: Number of different offsets along the z-th beam. 
  allocate1D( & lookupTable.numberOfOffsets, NUMBER_OF_OFFSET_ANGLES);
  
  // lookupTable.offsetDistIndex to the i-th index along the z-th beam.
  allocate2D( & lookupTable.offsetDistIndex, NUMBER_OF_OFFSET_ANGLES,
	      maxNumberOfDifferentSteps);
  
  /* For each cell precompute the distances along the beam. We only increment the step
   * if a new grid cell in the map is reached. Thus, for each angle and step we
   * also preprocess the resulting distance. */
  for ( int z = 0; z < NUMBER_OF_OFFSET_ANGLES; z++) {
    
    float xStep = cos( (float) z * angleStep) * distanceResolution;
    float yStep = sin( (float) z * angleStep) * distanceResolution;
    
    int step, offsetStep;
    
    lookupTable.xOffset[z][0] = 0;
    lookupTable.yOffset[z][0] = 0;
    lookupTable.offsetDistIndex[z][0] = 0;
    
    /* Store the indices in the map if going along the beam. */
    for ( step = 0, offsetStep = 0; step < maxDistanceIndex; step++) {
      
      int nextX = (int) ((float) (0.5 + step * xStep) / _resolution);
      int nextY = (int) ((float) (0.5 + step * yStep) / _resolution);
      
      /* Only increment the offset if a different cell of the map is hit! */
      if ( (nextX != lookupTable.xOffset[z][offsetStep]) ||
	   (nextY != lookupTable.yOffset[z][offsetStep])) {

	offsetStep++;

	lookupTable.xOffset[z][offsetStep] = nextX;
	lookupTable.yOffset[z][offsetStep] = nextY;
	lookupTable.offsetDistIndex[z][offsetStep] = step - 1;
      }
    }
    lookupTable.numberOfOffsets[z] = offsetStep + 1;
  }

  fprintf( stderr, "done.\n");

  lookupTable.initialized = true;
  
  return lookupTable;
}

//----------------------------------------------------------------------
// Determines the distance to the next wall given the current position and
// orientation of the beam.
//
// We precompute most of the indices so that searching for the next occupied cell
// in the beam direction will be done by table lookups.   
//----------------------------------------------------------------------
inline int
GridMap::expectedDistanceIndex( int beamStartX, int beamStartY, float orientation,
				distanceLookup& lookupTable)
{
  register int mapX, mapY;
  register float beamSum = 0.0;
  
  if ( ! lookupTable.initialized ) {
    cerr << "Error: cannot compute expected distance without initialized lookup table.\n";
    exit(0);
  }
  
  // Compute the index of the angle in the table.
  int lookupAngle = (int) (rad2Deg( normalizedAngle(orientation)) * lookupTable.angleToIndex)
    % lookupTable.numberOfAngles;

  
  /* Now step through the offsets and search for the first occupied cell. */
  for ( int step = 0; step < lookupTable.numberOfOffsets[lookupAngle]; step++) {

    mapX = lookupTable.xOffset[lookupAngle][step] + beamStartX;
    mapY = lookupTable.yOffset[lookupAngle][step] + beamStartY;
    
    if ( mapX >= 0 && mapX < _sizeX && mapY >= 0 && mapY < _sizeY)
	{
      
      if ( _prob[mapX][mapY] > OCCUPIED_THRESHOLD) 
	  {
		return lookupTable.offsetDistIndex[lookupAngle][step];
      }
    }
    else 
	{
      return lookupTable.offsetDistIndex[lookupAngle][step];
    }
  }
  return lookupTable.maxDistanceIndex - 1;
}



////////////////////////////////////////////////////////////////////////////////////////////////////
// writeExpectedDistanceTable
////////////////////////////////////////////////////////////////////////////////////////////////////
void GridMap::writeExpectedDistanceTable( string fileName )
{
  if ( ! _initialized ) {
    //cerr << "Error: grid map must be read BEFORE the expected dist table.\n";
    exit(0);
  }
  
  FILE* fp;
  if ( (fp = fopen( fileName.c_str(), "wb")) == NULL) {
    //cerr << "# Error: cannot open " << fileName << ".\n";
    return;
  }
  

  _distTabSizeX = sizeX();
  _distTabSizeY = sizeY();
  _distTabSizeRot = NUMBER_OF_ANGLES;

  fwrite(&_distTabSizeX, sizeof(int), 1, fp );
  fwrite(&_distTabSizeY, sizeof(int), 1, fp );
  fwrite(&_distTabSizeRot, sizeof(int), 1, fp );
  fwrite(&_distanceResolution, sizeof(double), 1, fp );
    
  //buffer the expecteddistance indexes:
  /* Allocate memory and get the data. */
  allocate3D( &_expectedDistTable, sizeX(), sizeY(), NUMBER_OF_ANGLES);
    
  for ( int x = 0; x < _distTabSizeX; x++)
  {
    for ( int y = 0; y < _distTabSizeY; y++) 
    {
      for ( int z = 0; z < NUMBER_OF_ANGLES; ++z )
      {
        float index = expectedDistanceIndex(x,y,deg2Rad(z),_onlineLookup);
        float orientation_rt_deg = z+(CONE_SIZE_DEGREES/2);
        float orientation_lft_deg = z-(CONE_SIZE_DEGREES/2);

        float index_rt = expectedDistanceIndex( x, y, deg2Rad(orientation_rt_deg), _onlineLookup );
        float index_lft = expectedDistanceIndex( x, y, deg2Rad(orientation_lft_deg), _onlineLookup );

        float variance = util_GetDeviation( index_lft, index, index_rt);

        unsigned char c = expectedDistanceIndex(x,y,deg2Rad(z),_onlineLookup);
        
        if( variance >= DEVIATION_THRESHOLD_CONE )
        {
          c+=128;
        }

        _expectedDistTable[x][y][z] = c;

      }
      
      }
  }

  
  for ( int x2 = 0; x2 < sizeX(); x2++)
  {
    for ( int y2 = 0; y2 < sizeY(); y2++) 
    {
      fwrite( _expectedDistTable[x2][y2], sizeof(unsigned char), NUMBER_OF_ANGLES, fp); 
    }
  }


  _distance_table_initialized = true;
  fclose( fp);
}

void GridMap::readExpectedDistanceTable( string fileName )
{
  if ( ! _initialized ) {
    //cerr << "Error: grid map must be read BEFORE the expected dist table.\n";
    exit(0);
  }

  FILE* fp;
  if ( (fp = fopen( fileName.c_str(), "rb")) == NULL) {
    //cerr << "# Error: cannot open " << fileName << ".\n";
    return;
  }

  fread(&_distTabSizeX, sizeof(int), 1, fp );
  fread(&_distTabSizeY, sizeof(int), 1, fp );
  fread(&_distTabSizeRot, sizeof(int), 1, fp );
  fread(&_distanceResolution, sizeof(double), 1, fp );
  
  /* Allocate memory and get the data. */
  allocate3D( &_expectedDistTable, sizeX(), sizeY(), NUMBER_OF_ANGLES );
    
  for ( int x = 0; x < sizeX(); x++)
  {
    for ( int y = 0; y < sizeY(); y++) 
    {
      fread( _expectedDistTable[x][y], sizeof(unsigned char), NUMBER_OF_ANGLES, fp);
    }
  }

  fclose( fp);

  _distance_table_initialized = true;
}


//----------------------------------------------------------------------
inline indexGridPosition
GridMap::indexPos( float x, float y) const
{
  indexGridPosition pos;

  if ( x < _offsetX)
    pos.x = -1;
  else
    pos.x = (int) ((x - _offsetX) / _resolution);

  if ( y < _offsetY)
    pos.y = -1;
  else
    pos.y = (int) ((y - _offsetY) / _resolution);

  return pos;
}

////////////////////////////////////////////////////////////////////////////
////////////////////////////////////////////////////////////////////////////
//
// Initialization and destructors of the grid map class.
//
////////////////////////////////////////////////////////////////////////////
////////////////////////////////////////////////////////////////////////////


//----------------------------------------------------------------------
 bool
GridMap::initialized() const
{
  return _initialized;
}

//----------------------------------------------------------------------
 string
GridMap::fileName() const
{
  return _fileName;
}

//----------------------------------------------------------------------
 float
GridMap::minX() const
{
  return _offsetX;
}

//----------------------------------------------------------------------
 float
GridMap::minY() const
{
  return _offsetY;
}


//----------------------------------------------------------------------
 float
GridMap::maxX() const
{
  return _maxRealX;
}

//----------------------------------------------------------------------
 float
GridMap::maxY() const
{
  return _maxRealY;
}

//----------------------------------------------------------------------
 int
GridMap::sizeX() const
{
  return _sizeX;
}

//----------------------------------------------------------------------
 int
GridMap::sizeY() const
{
  return _sizeY;
}

//----------------------------------------------------------------------
 float
GridMap::resolution() const
{
  return _resolution;
}

 //----------------------------------------------------------------------
bool
GridMap::inMap( int x, int y) const
{
  return x >= 0 && y >= 0 && x < _sizeX && y < _sizeY;
}

//----------------------------------------------------------------------
bool
GridMap::inMap( const indexGridPosition & pos) const
{
  return pos.x >= 0 && pos.y >= 0 && pos.x < _sizeX && pos.y < _sizeY;
}

bool
GridMap::occupied( float Xmm, float Ymm ) const
{

  float Xcm = Xmm / MILLIMETERS_PER_CENTIMETER;
  float Ycm = Ymm / MILLIMETERS_PER_CENTIMETER;

  indexGridPosition pos = indexPos( Xcm, Ycm );
  if ( _prob[pos.x][pos.y] > OCCUPIED_THRESHOLD || _prob[pos.x][pos.y] == -1 )
    return true;
  else
    return false;
}

float util_GetDeviation( float indexleft, float index, float indexright )
{

   float mean = (index + indexleft + indexright) / 3;
   float variance = (pow(index-mean,2)+pow(indexleft-mean,2)+pow(indexright-mean,2))/3;
   return variance;
}

#ifdef TEST_GRID
GridMap* grid=new GridMap("uw-new.map");
#endif

