#ifndef GRIDMAP_LOADED
#define GRIDMAP_LOADED

#include "State.h"
#include "Saphira.h"
#include <utility>

using namespace std;

//CONSTANTS:
const float DEFAULT_RESOLUTION   = 15.0;
const float DEFAULT_DISTANCE_RES_LOOKUP = 3.0;
const float DEFAULT_UNKNOWN_PROB = -1.0;
const float DEFAULT_OUTSIDE_PROB =  1.0;
const float P_OBSTACLE_THRESHOLD = 0.5;
const float CENTIMETERS_PER_PIXEL = 1.0;
const float MILLIMETERS_PER_CENTIMETER = 10.0;

#define NUMBER_OF_ANGLES 360
#define OCCUPIED_THRESHOLD 0.1
#define NUMBER_OF_OFFSET_ANGLES 360
#define ANGLE_TO_INDEX 1.0
#define CONE_SIZE_DEGREES 30.0
#define DEVIATION_THRESHOLD_CONE 20.0
#define DEFAULT_MAX_DISTANCE_INDEX 128

typedef struct{
  float x;
  float y;
} realGridPosition;

typedef struct{
  int x;
  int y;
} indexGridPosition;

// Look up tables to compute the expected distance to a wall.
typedef struct{
  int   maxDistanceIndex;
  float distanceResolution;
  bool  initialized;
  int** xOffset;
  int** yOffset;
  int*  numberOfOffsets;
  int** offsetDistIndex;
  float angleToIndex;
  int   numberOfAngles;
} distanceLookup;
  
//Returns an integer representing the expected distance in millimeters, and a
//bool indicating whether or not the scan for this beam is "good" or "bad", where
//"bad" corresponds to a beam that hit at a bad angle.
typedef pair< int, bool > expectedDistanceRet;

/////////////////////////////////////////////////////////////////
//  Occupancy grid map.
/////////////////////////////////////////////////////////////////
class GridMap : public SfArtifact {

public:

  ~GridMap();

  // Initializations.
  GridMap( string mapFile  );

  GridMap( string mapFile, string distFile );

  // Reallocates memory! Assumes that operator= is NOT overridden!
  // If you want the memory to be shared, then use the = operator.
  // Otherwise use initialize operator or later: dest = (GridMap) src;
  // This will invoke the copy operator.
  //GridMap( const GridMap & map);
  
  string          fileName() const;


  int      sizeX() const;
  int      sizeY() const;
  float    resolution() const;
  
  //----------------------------------------------------------------------
  // Access functions.
  //----------------------------------------------------------------------
  bool          initialized() const;

  // Computes the distance to the next wall given the start position of the beam.
  // Distance is given in millimeters.
  expectedDistanceRet  expectedDistance( int BotX, int BotY, float BotOrientation, int SensorNumber);

  float         minX() const;
  float         minY() const;
  float         maxX() const;
  float         maxY() const;
  realGridPosition  realPos( int x, int y) const;
  indexGridPosition indexPos( float x, float y) const;
  bool inMap( int x, int y) const;
  bool inMap( const indexGridPosition & pos) const;
  bool occupied( float Xmm,  float Ymm ) const;

  //----------------------------------------------------------------------
  // Display stuff.
  //----------------------------------------------------------------------
  //We need to override and define this function so the GridMap can be drawn
  //in Saphira.
  void          draw( SfWin *w );

  double        mapToDisplay( double mapValue );
  double        displayToMap( double displayValue);

protected:
  // Read in the table... not an option for this GridMap class.
  bool            read( string& mapFile);
public:
  //test & distance file stuff. but maybe people will want to use it anyway...
  void            writeExpectedDistanceTable( string  filename );
  void            readExpectedDistanceTable( string  filename );
  int             testExpectedDistanceIndex( int x, int y, float th )
  { return expectedDistanceIndex( x, y, th, _onlineLookup);}
protected:
  // Tables needed for fast computation of the expected distances.
  distanceLookup precomputedLookupTables( float distanceResolution = DEFAULT_DISTANCE_RES_LOOKUP,
    int maxDistanceIndex = DEFAULT_MAX_DISTANCE_INDEX );

  // Computes the distance to the next wall given the start position of the beam.
  // Distance is given in indices. This is for computing the distance on the fly, and should not be
  // used over the precomputed table (which will be much faster).  The index is in the distance resolution
  // of the distanceLookup table (the distanceResolution default is 3.0 cm).
  inline int  expectedDistanceIndex( int beamStartX, int beamStartY, float orientation,
				     distanceLookup& lookupTable);

  bool          _initialized;
  bool          _distance_table_initialized;
  string        _fileName;
  int           _sizeX;
  int           _sizeY;
  float         _resolution;
  float         _offsetX;
  float         _offsetY;
  float         _maxRealX;
  float         _maxRealY;
  float         _unknownProb;
  float**       _prob;
  double        _distanceResolution;

  unsigned char***  _expectedDistTable;     // (expected dist[l])
  int               _distTabSizeX;
  int               _distTabSizeY;
  int               _distTabSizeRot;
  distanceLookup _onlineLookup;
};

#endif