
/***************************************************************************
 ***************************************************************************
 ***************************************************************************
 *****
 *****
 ***** This file is part of the robot control software for the UW AIBO
 ***** Sony RoboCup league.
 *****
 ***** This software is confidential.
 *****
 ***************************************************************************
 ***************************************************************************
 ***************************************************************************
 ***** 
 ***** THERE IS NO WARRANTY FOR THE PROGRAM, TO THE EXTENT PERMITTED
 ***** BY APPLICABLE LAW.  EXCEPT WHEN OTHERWISE STATED IN WRITING
 ***** THE COPYRIGHT HOLDERS AND/OR OTHER PARTIES PROVIDE THE PROGRAM
 ***** "AS IS" WITHOUT WARRANTY OF ANY KIND, EITHER EXPRESSED OR
 ***** IMPLIED, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES
 ***** OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE.  THE
 ***** ENTIRE RISK AS TO THE QUALITY AND PERFORMANCE OF THE PROGRAM
 ***** IS WITH YOU.  SHOULD THE PROGRAM PROVE DEFECTIVE, YOU ASSUME
 ***** THE COST OF ALL NECESSARY SERVICING, REPAIR OR CORRECTION.
 ***** 
 ***** IN NO EVENT UNLESS REQUIRED BY APPLICABLE LAW OR AGREED TO IN
 ***** WRITING WILL ANY COPYRIGHT HOLDER, OR ANY OTHER PARTY WHO MAY
 ***** MODIFY AND/OR REDISTRIBUTE THE PROGRAM AS PERMITTED ABOVE, BE
 ***** LIABLE TO YOU FOR DAMAGES, INCLUDING ANY GENERAL, SPECIAL,
 ***** INCIDENTAL OR CONSEQUENTIAL DAMAGES ARISING OUT OF THE USE OR
 ***** INABILITY TO USE THE PROGRAM (INCLUDING BUT NOT LIMITED TO
 ***** LOSS OF DATA OR DATA BEING RENDERED INACCURATE OR LOSSES
 ***** SUSTAINED BY YOU OR THIRD PARTIES OR A FAILURE OF THE PROGRAM
 ***** TO OPERATE WITH ANY OTHER PROGRAMS OR FAILURE TO CONTROL A
 ***** PHYSICAL DEVICE OF ANY TYPE), EVEN IF SUCH HOLDER OR OTHER
 ***** PARTY HAS BEEN ADVISED OF THE POSSIBILITY OF SUCH DAMAGES.
 *****
 ***************************************************************************
 ***************************************************************************
 ***************************************************************************
 *****
 ***** Source file:     $Source: /home/henri/CVS/aibo-2001/src/tools/src/aibo/aibo.c,v $
 *****
 ***** Created by:      $Author: fox $
 *****
 ***** Revision #:      $Revision: 1.9 $
 *****
 ***** Date of revision $Date: 2001/04/23 02:49:59 $
 *****
 ***************************************************************************
 ***************************************************************************
 ***************************************************************************
 *
 *                  ----- REVISION HISTORY -----
 *
 * $Log: aibo.c,v $
 *
 *
 *
 ***************************************************************************
 ***************************************************************************
 ***************************************************************************/

#include "general.h"
#include "State.h"
#include "function.h"
#include "function.icc"


////////////////////////////////////////////////////////////////////////////
////////////////////////////////////////////////////////////////////////////
//
// Initialization and destructors of State class.
//
////////////////////////////////////////////////////////////////////////////
////////////////////////////////////////////////////////////////////////////

//----------------------------------------------------------------------
//----------------------------------------------------------------------

State::State() 
{
  if ( SAMPLE_TEST)
    cerr << "State " << NUMBER_OF_DIMENSIONS << "\n";
}


//----------------------------------------------------------------------
// Automatically sets dimension to three.
//----------------------------------------------------------------------
State::State( float x, float y, float orientation)
{
  if ( SAMPLE_TEST)
    cerr << "Set state directly.\n";
  
  _value[X_DIM] = x;
  _value[Y_DIM] = y;
  _value[ROT_DIM] = orientation;
}

//----------------------------------------------------------------------
// Copy operator.
//----------------------------------------------------------------------

State::State( const State & s) 
{
  for ( int dim = 0; dim < NUMBER_OF_DIMENSIONS; dim++)
    _value[dim] = s.value(dim);
}


//----------------------------------------------------------------------
// Copy operator.
//----------------------------------------------------------------------
 State &
State::operator = (const State & s)
{
  // Check whether dimension is different.
  if ( SAMPLE_TEST)
    cerr << "State copy without allocation\n";
  
  // Copy the values.
  for ( int dim = 0; dim < dimension(); dim++)
    _value[dim] = s.value(dim);
  
  return *this;
}


//----------------------------------------------------------------------
// Free memory if necessary.
//----------------------------------------------------------------------

State::~State()
{
  if ( SAMPLE_TEST)
    cerr << "Destroy " << dimension() << "-d state at " << _value << " ... ";
}


//----------------------------------------------------------------------
// Comparison operator.
//----------------------------------------------------------------------
 bool
State::operator != ( const State& s) const
{
  for ( int dim = 0; dim < dimension(); dim++)
    if ( value(dim) != s.value(dim))
      return true;
  return false;
}

//----------------------------------------------------------------------
// Comparison operator.
//----------------------------------------------------------------------
 bool
State::operator == ( const State& s) const
{
  for ( int dim = 0; dim < dimension(); dim++)
    if ( value(dim) != s.value(dim))
      return false;
  return true;
}

//----------------------------------------------------------------------
// Substracts the states in each dimension.
//----------------------------------------------------------------------
 State
State::operator - ( const State& s) const
{
  State diff = *this;
  
  for ( int dim = 0; dim < dimension(); dim++)
    diff[dim] -= s.value(dim);
  return diff;
}

//----------------------------------------------------------------------
// Substracts the states in each dimension.
//----------------------------------------------------------------------
 State
State::operator + ( const State& s) const
{
  State add = *this;
  
  for ( int dim = 0; dim < dimension(); dim++)
    add[dim] += s.value(dim);
  return add;
}

//----------------------------------------------------------------------
// Read access to a state value.
//----------------------------------------------------------------------
 float&
State::operator[]( int dim)
{
#ifdef RANGE_CHECKING
  rangeCheck( dim, dimension(), "State::[]: ");
#endif
  
  return _value[dim];
}


/*
//----------------------------------------------------------------------
// output of a State structure in gnuplot format 
//----------------------------------------------------------------------
 ostream&
operator<<( ostream& os, const State& state)
{
  if (  SAMPLE_TEST)
    cerr << "STATE: ";
  for ( int dim = 0; dim < state.dimension(); dim++) {
    if ( dim > 0) os << " ";
    os << state.value(dim);
  }
  return os;
}
*/

//----------------------------------------------------------------------
// We assume the noise to be Gaussian and not correlated among the dimensions.
//----------------------------------------------------------------------
 void
State::addNoise( const State& stdDev)
{
  for ( int dim = 0; dim < dimension(); dim++)
    _value[dim] += stdDev.value(dim) * randomGauss();
}

//----------------------------------------------------------------------
// Checks whether the state is smaller than the other state.
//----------------------------------------------------------------------
 bool
State::smallerOrEqual( const State& s) const
{
  for ( int dim = 0; dim < dimension(); dim++)
    if ( value(dim) > s.value(dim))
      return false;

  return true;	
}

//----------------------------------------------------------------------
// Checks whether the state is larger than the other state.
//----------------------------------------------------------------------
 bool
State::largerOrEqual( const State& s) const
{
  for ( int dim = 0; dim < dimension(); dim++)
    if ( value(dim) < s.value(dim))
      return false;

  return true;	
}

//----------------------------------------------------------------------
// Normalize the angle between 0 and 2pi.
//----------------------------------------------------------------------
 void
State::normalizeOrientation()
{
  _value[ROT_DIM] = normalizedAngle( _value[ROT_DIM]);
}

//----------------------------------------------------------------------
// Checks whether the state is in the cube defined by the two corners.
//----------------------------------------------------------------------
 bool
State::inBoundingBox( const State& minCorner, const State& maxCorner) const
{
  return largerOrEqual( minCorner) && smallerOrEqual( maxCorner);
}


//----------------------------------------------------------------------
//----------------------------------------------------------------------
 int
State::dimension() const
{
  return NUMBER_OF_DIMENSIONS;
}

//----------------------------------------------------------------------
//----------------------------------------------------------------------
 const float&
State::value( int dim) const
{
#ifdef RANGE_CHECKING
  rangeCheck( dim, dimension(), "State::[]: ");
#endif
  return _value[dim];
}

//----------------------------------------------------------------------
// Treat the state as if it would represent the edges of a hyper cube.
//----------------------------------------------------------------------
 float
State::volume() const
{
  float vol = 1.0;
  
  for ( int dim = 0; dim < dimension(); dim++)
    vol *= _value[dim];

  return vol;
}

//----------------------------------------------------------------------
// Treat the state as if it would represent a vector.
//----------------------------------------------------------------------
 float
State::length() const
{
  float length = sqr( _value[X_DIM]) + sqr( _value[Y_DIM]);

  return sqrt(length);
}


//----------------------------------------------------------------------
// Returns a state drawn uniformly between min and max.
//----------------------------------------------------------------------
 void
State::setRandom( const State& min, const State& max)
{
  for ( int dim = 0; dim < dimension(); dim++) {
    
    float minVal = min.value(dim);
    float size   = max.value(dim) - minVal;
    _value[dim]  = minVal + randomZeroToOne() * size;
  }
}

